# Copyright (c) 2009 Mathias Kaerlev.
# See LICENSE for details.

# Copyright (c) 2009 Mathias Kaerlev.
# See LICENSE for details.

import traceback

from twisted.protocols.basic import LineOnlyReceiver

from pySndo.cipher import MMFBlowfish
from pySndo.network.game import Message
from pySndo.network.info import ClientRequest, ServerResponse, infoKey
from pySndo.game.team import Police, Terrorist

bugLines = ['GET', 'From']

def lineBugged(line):
    matches = [buggedLine for buggedLine in bugLines 
               if line.startswith(buggedLine)]
    return len(matches) > 0

class GameProtocol(LineOnlyReceiver):
    """
    The base game protocol for client and server
    
    @ivar name: Name of the client
    @ivar id: ID of the client
    @ivar team: The team of this player.
    @type team: L{Terrorist} or L{Police} or None (if no team yet)
    """
    name = None
    id = None
    team = None
    isAccepted = False
    
    # implementation-specific methods
    
    def lineReceived(self, line):
        try:
            decrypted = self.factory.cipher.decrypt(line)
            # BUG: in some cases, the MMF server might errornously send us
            # webserver queries, so ignore them
            if self._isClient and lineBugged(line):
                return
            newPacket = Message(decrypted, self._isClient)
            if self._isClient and newPacket.way in (0, 3):
                return
            self.packetReceived(newPacket)
        except Exception, e:
            if 'decrypted' in locals():
                realData = decrypted
            else:
                realData = line
            self.packetCorrupted(realData, e)
    
    def sendLine(self, line):
        LineOnlyReceiver.sendLine(self, self.factory.cipher.encrypt(line))
    
    # action methods
    
    def sendType(self, packetType):
        """
        Send a packetloader to the other end.
        
        @type packetType: See L{pySndo.network.packetloaders}
        """
    
    # implementation overridables

    def packetCorrupted(self, data, e):
        """
        Called if an exception was raised during parsing/processing
        @param data: Raw data of received packet.
        @type data: str or None
        @param e: Exception raised
        """
        traceback.print_exc()
        file = open('./bug.txt', 'ab')
        file.writelines(['** corrupted: %r **' % data])
        traceback.print_exc(file = file)
        raise e
    
    def packetReceived(self, packet):
        """
        Called upon receiving a protocol packet.
        
        @param packet: Received packet
        @type packet: Message instance
        """
        
    # common overridables
    
    def crateSpawned(self, crate):
        """
        Called when a new crate has been spawned on the map.
        
        @type crate: L{GunCrate}, L{GrenadeCrate} or L{HealthCrate} object
        """
        
    def crateDestroyed(self, crate):
        """
        Called when a crate has exceeded its max life time.
        
        @type crate: L{GunCrate}, L{GrenadeCrate} or L{HealthCrate} object
        """


class GameFactory:
    """
    The base game factory for both client and server.
    
    @ivar connections: A list of L{pySndo.network.server.ServerGameProtocol} (if server) 
        or L{pySndo.network.client.OtherClient} (if client) instances.
    
    @ivar cipher: The cipher object used to encrypt 
        networking data.
    @ivar key: The key used to encrypt networking data.
    
    @ivar police: A L{Police} object representing the 
        police team.
    @ivar terrorist: A L{Terrorist} object representing 
        the terrorist team.
    
    @ivar map: The current map.
    @type map: L{Map} instance
    
    @ivar crates: Crates currently spawned on the map.
    @type crates: List of L{GunCrate}, L{GrenadeCrate} 
        and L{HealthCrate} objects
    
    @ivar bombedAreas: Bombed areas on map
    @type bombedAreas: List of L{BombedArea}
    """
    connections = None

    key = None
    cipher = None

    police = None
    terrorist = None
    
    crates = None
    
    def startFactory(self, *arg, **kw):
        self.cipher = MMFBlowfish(self.key)
        self.connections = []
        self.bombedAreas = []
        self.crates = []
        self.police = Police(self)
        self.terrorist = Terrorist(self)
        
    def playerFromName(self, playerName):
        return [connection for connection
                in self.connections
                if connection.name == playerName][0]

    def playerFromId(self, playerId):
        playerMatches = [connection for connection
                         in self.connections
                         if connection.id == playerId]
        if not playerMatches:
            raise IndexError('player %s could not be found' % playerId)
        return playerMatches[0]
        
    def aliveTeamNumbers(self):
        """
        Return the number of alive players on each
        team
        
        @return: 2-item tuple (Terrorist, Police)
        """
        return (len(self.terrorist.getAliveConnections()), 
                len(self.police.getAliveConnections()))
        
    def teamNumbers(self):
        """
        Return the number of players on each
        team
        
        @return: 2-item tuple (Terrorist, Police)
        """
        return len(self.police.getConnections()), len(self.terrorist.getConnections())
        
    def teamFromName(self, teamName):
        if teamName == 'Police':
            return self.police
        elif teamName == 'Terrorist':
            return self.terrorist
        else:
            return None
            
    def teamFromObject(self, team):
        if team == self.police:
            return 'Police'
        elif team == self.terrorist:
            return 'Terrorist'

class InfoProtocol(LineOnlyReceiver):
    """
    The base info protocol
    """
    
    def lineReceived(self, line):
        try:
            if self._isClient:
                # BUG: in some cases, the MMF server might errornously send us
                # webserver queries, so ignore them
                if lineBugged(line):
                    return
                decrypted = self.factory.cipher.decrypt(line)
                newPacket = ServerResponse(decrypted)
            else:
                newPacket = ClientRequest(line)
            self.packetReceived(newPacket)
        except Exception, e:
            if 'decrypted' in locals():
                realData = decrypted
            else:
                realData = line
            self.packetCorrupted(realData, e)

    # implementation overridables

    def packetCorrupted(self, data, e):
        """
        Called if an exception was raised during parsing/processing
        @param data: Raw data of received packet.
        @type data: str or None
        @param e: Exception raised
        """
        traceback.print_exc()
        traceback.print_exc(file = open('./bug.txt', 'ab'))
        raise e

    def packetReceived(self, packet):
        """
        Called upon receiving a protocol packet.
        
        @param packet: Received packet
        @type packet: L{ClientRequest} or L{ServerResponse} instance
        """
        
    def sendType(self, loader):
        if self._isClient:
            newPacket = ClientRequest()
        else:
            newPacket = ServerResponse()

        newPacket.type = loader
        self.sendLine(newPacket.generate())
            
        
    def sendLine(self, line):
        if self._isClient:
            LineOnlyReceiver.sendLine(self, line)
        else:
            LineOnlyReceiver.sendLine(self, self.factory.cipher.encrypt(line))
        
    
class InfoFactory:
    """
    The base info factory

    @ivar cipher: The cipher object used to encrypt/decrypt
        data to/from the other end.
    """
    cipher = None

    def startFactory(self, *arg, **kw):
        self.cipher = MMFBlowfish(infoKey)
